package ru.dbotthepony.kstarbound.server

import ru.dbotthepony.kstarbound.client.network.packets.InitialChunkDataPacket
import ru.dbotthepony.kstarbound.client.network.packets.JoinWorldPacket
import ru.dbotthepony.kstarbound.server.network.ServerChannels
import ru.dbotthepony.kstarbound.server.network.ServerPlayer
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.util.MailboxExecutorService
import java.io.Closeable
import java.io.File
import java.util.Collections
import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.LockSupport
import java.util.concurrent.locks.ReentrantLock
import kotlin.concurrent.withLock

abstract class StarboundServer(val root: File) : Closeable {
	init {
		if (!root.exists()) {
			check(root.mkdirs()) { "Unable to create ${root.absolutePath}" }
		} else if (!root.isDirectory) {
			throw IllegalArgumentException("${root.absolutePath} is not a directory")
		}
	}

	val worlds: MutableList<ServerWorld> = Collections.synchronizedList(ArrayList<ServerWorld>())

	val thread = Thread(::runThread, "Starbound Server ${threadCounter.incrementAndGet()}")
	val mailbox = MailboxExecutorService(thread)

	val settings = ServerSettings()
	val channels = ServerChannels(this)
	val lock = ReentrantLock()

	@Volatile
	var isClosed = false
		private set

	init {
		thread.isDaemon = true
		thread.start()
	}

	fun playerInGame(player: ServerPlayer) {
		val world = worlds.first()
		player.world = world
		world.players.add(player)
		player.connection.send(JoinWorldPacket(world))
	}

	protected abstract fun close0()

	private fun runThread() {
		while (!isClosed) {
			mailbox.executeQueuedTasks()
			LockSupport.park()
		}
	}

	final override fun close() {
		lock.withLock {
			if (isClosed) return

			channels.close()
			worlds.forEach { it.close() }
			close0()
			LockSupport.unpark(thread)
		}
	}

	companion object {
		private val threadCounter = AtomicInteger()
	}
}
